/************************************************************
 *
 * Copyright (c) 2001 Chemi. All rights reserved.
 * 
 * This program and the accompanying materials
 * are made available under the terms of the MIT License
 * which accompanies this distribution, and is available at
 * http://www.opensource.org/licenses/mit-license.html
 *
 ************************************************************/

package es.org.chemi.games.minesweeper.core;

import es.org.chemi.games.minesweeper.MinesweeperPlugin;
import es.org.chemi.games.minesweeper.events.CellListener;
import es.org.chemi.games.minesweeper.util.Preferences;
import es.org.chemi.games.minesweeper.ui.Cell;
import es.org.chemi.games.minesweeper.ui.MainView;
import es.org.chemi.games.minesweeper.ui.NewRecordDialog;
import es.org.chemi.games.util.Constants;
import es.org.chemi.games.util.ResourceManager;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Vector;
import org.eclipse.core.runtime.Platform;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;

public class Controller 
{
	private Composite minesField = null;
	private MainView view = null;
	private CellListener cellListener = null;
	
	private int[] mines = null;
	private String traceEvents = Platform.getDebugOption("es.org.chemi.games.minesweeper/debug/events");
	
	private Preferences preferences = null;
	private Hashtable records = null;
	
	private boolean sw = false; // Used to detect when the two mouse buttons have been pressed.
	
	public Controller(MainView view, Preferences preferences, Hashtable records)
	{
		this.view = view;
		this.preferences = preferences;
		this.records = records;
		this.mines = new int[preferences.getMinesCount()];		
		this.cellListener = new CellListener(view);
	}
	
	// Getters
	public Preferences getPreferences()
	{
		return this.preferences;
	}

	public Hashtable getRecords()
	{
		return this.records;
	}
	
	public void changeMode(String mode)
	{
		this.preferences.setMode(mode);
	}
	
	public void refresh(int minesCount, int rows, int columns)
	{
		preferences.setMinesCount(minesCount);
		preferences.setRows(rows);
		preferences.setColumns(columns);
		
		mines = new int[preferences.getMinesCount()];
		view.getCounter().reset(preferences.getMinesCount());
		
		GridLayout layout = new GridLayout();
		layout.numColumns = preferences.getColumns();
		layout.marginWidth = 0;
		layout.marginHeight = 0;
		layout.horizontalSpacing = 0;
		layout.verticalSpacing = 0;
		layout.makeColumnsEqualWidth = true;
		minesField.setLayout(layout);

		initMinesField();
	}
	
	// Initialize the game.
	public void initMinesField(Composite minesField)
	{
		this.minesField = minesField;
		
		this.fillComposite();
		this.fillMines();
		this.fillWarnings();
	}
	
	// Reinitialize the game.
	public void initMinesField()
	{
		if(view.getTimer().isRunning())
		{
			view.getTimer().stop();					
			view.getTimer().resetRunning();
		}
		else
			view.getTimer().resetStopped();

		this.initMinesField(minesField);
		view.getMainButton().setImage(ResourceManager.getResourceManager().getImage(Constants.IMAGE_SMILE));		
	}
	
	// Fill with cells.
	private void fillComposite()
	{
		// Tracing.
		if(MinesweeperPlugin.getDefault().isDebugging())
			System.out.println("[Minesweeper - " + Constants.TRACE_INFO + "] - " + Calendar.getInstance().getTime() + " - Creating mines field...");

		boolean sw = false;
		
		// Check if minesField needs to be repainted or just refreshed.
		if(minesField.getChildren().length == preferences.getRows() * preferences.getColumns())
		{
			Control[] c = minesField.getChildren();
			Cell cell = null;
			CellInfo cellInfo = null;
			
			for(int i=0; i<c.length; i++)
			{
				cell = (Cell)c[i];
				cellInfo = (CellInfo)cell.getData();
				
				if(cell.isLabel())
				{
					cell.setType(Constants.TYPE_BUTTON);
					cell.setImage(null);
					cell.setBackground(ResourceManager.getResourceManager().getColor(Constants.COLOR_PLATFORM));
					cell.redraw();
				}
				if(cellInfo.isFlag())
				{
					cellInfo.setFlag(false);
					cell.setImage(null);
					cell.redraw();					
				}
				if(cellInfo.isQuestion())
				{
					cellInfo.setQuestion(false);
					cell.setImage(null);
					cell.redraw();					
				}
				if(cellInfo.isMine())
					cellInfo.setMine(false);
					
				cellInfo.setWarning(0);
				cell.removeMouseListener(this.cellListener);				
				cell.addMouseListener(this.cellListener);
			}

			return;
		}
		
		if(minesField.getChildren().length != 0)
		{
			Control[] c = minesField.getChildren();
			for(int i=0; i<c.length; i++)
				c[i].dispose();
			sw = true;
		}
		
		Cell cell = null;
		GridData data = null;
		CellInfo cellInfo = null;
		
		for(int i=0;i<preferences.getRows();i++)
		{
			for (int j=0;j<preferences.getColumns();j++)
			{
				cell = new Cell(minesField,SWT.PUSH);
				data = new GridData();
				data.grabExcessHorizontalSpace = true;
				data.grabExcessVerticalSpace = true;
				data.horizontalAlignment = GridData.FILL;
				data.verticalAlignment = GridData.FILL;
				cell.setLayoutData(data);
				cell.addMouseListener(this.cellListener);
				cellInfo = new CellInfo(i,j);
				if(i == 0)
					cellInfo.setBorder(Constants.TOP_BORDER);
				else if(i == preferences.getRows()-1)
					cellInfo.setBorder(Constants.DOWN_BORDER);
				if(j == 0)
					cellInfo.setBorder(Constants.LEFT_BORDER);
				else if(j == preferences.getColumns()-1)
					cellInfo.setBorder(Constants.RIGHT_BORDER);
				cell.setData(cellInfo);
			}
		}
		
		// Check if a layout is needed.
		if(sw)
			minesField.layout(true);
	}
	
	// Fill with mines.
	private void fillMines()
	{
		// Tracing.
		if(MinesweeperPlugin.getDefault().isDebugging())
			System.out.println("[Minesweeper - " + Constants.TRACE_INFO + "] - " + Calendar.getInstance().getTime() + " - Filling with mines the field...");
		
		int aux = 0;
		CellInfo cellInfo = null;
		Control[] c = minesField.getChildren();
	
		// Calculate the mines.	
		while(aux < preferences.getMinesCount())
		{
			int index = (int)Math.rint(Math.random() * (preferences.getColumns() * preferences.getRows() -1));
			cellInfo = (CellInfo)c[index].getData();
			if(!cellInfo.isMine())
			{
				cellInfo.setMine(true);
				mines[aux] = index;
				aux++;
			}
		}		
	}
	
	// Calculate the number of mines beside a specific cell.
	private void fillWarnings()
	{
		// Tracing.
		if(MinesweeperPlugin.getDefault().isDebugging())
			System.out.println("[Minesweeper - " + Constants.TRACE_INFO + "] - " + Calendar.getInstance().getTime() + " - Filling with warnings the field...");

		// Variables.
		CellInfo cellInfo = null;
		CellInfo aux = null;
		int x = 0;
		int y = 0;
		int columns = preferences.getColumns();
		
		Control[] c = minesField.getChildren();
		for(int i=0;i<mines.length;i++)
		{
			cellInfo = (CellInfo)c[mines[i]].getData();
			
			// Algorithm
			if(cellInfo.isTopBorder())
			{
				if(cellInfo.isLeftBorder())
				{
					// The same line.
					aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);
						
					// The line below.
					aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);						
				}
				else if(cellInfo.isRightBorder())
				{
					// The same line.
					aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					
					// The line below.
					aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y - 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y].getData();
					aux.setWarning(aux.getWarning() + 1);												
				}
				else
				{
					// The same line.
					aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);
						
					// The line below.
					aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y - 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);												
				}
			}
			else if(cellInfo.isDownBorder())
			{
				if(cellInfo.isLeftBorder())
				{
					// The same line.
					aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					
					// The line above.
					aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);						
				}
				else if(cellInfo.isRightBorder())
				{
					// The same line.
					aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					
					// The line above.
					aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y - 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y].getData();
					aux.setWarning(aux.getWarning() + 1);												
				}
				else
				{
					// The same line.
					aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);
						
					// The line above.
					aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y - 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);												
				}
			}
			else if(cellInfo.isLeftBorder())
			{
					// The same line.
					aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);
					
					// The line above.
					aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);						

					// The line below.
					aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
					aux.setWarning(aux.getWarning() + 1);
					aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
					aux.setWarning(aux.getWarning() + 1);											
			}
			else if(cellInfo.isRightBorder())
			{
				// The same line.
				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				aux.setWarning(aux.getWarning() + 1);
						
				// The line above.
				aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
				aux.setWarning(aux.getWarning() + 1);
				aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y - 1].getData();
				aux.setWarning(aux.getWarning() + 1);						

				// The line below.
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
				aux.setWarning(aux.getWarning() + 1);
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y - 1].getData();
				aux.setWarning(aux.getWarning() + 1);																
			}
			else  // It doesn't touch any border.
			{
				// The same line.
				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				aux.setWarning(aux.getWarning() + 1);
				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
				aux.setWarning(aux.getWarning() + 1);
						
				// The line above.
				aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
				aux.setWarning(aux.getWarning() + 1);
				aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y - 1].getData();
				aux.setWarning(aux.getWarning() + 1);						
				aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
				aux.setWarning(aux.getWarning() + 1);						

				// The line below.
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
				aux.setWarning(aux.getWarning() + 1);
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y - 1].getData();
				aux.setWarning(aux.getWarning() + 1);																
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
				aux.setWarning(aux.getWarning() + 1);																						
			}
		}
	}
	
	public void showAllMines()
	{
		CellInfo cellInfo = null;
		Cell cell = null;
		GridData data = null;
		Control[] controls = minesField.getChildren();
		
		// Tracing.
		if(MinesweeperPlugin.getDefault().isDebugging())
			System.out.println("[Minesweeper - " + Constants.TRACE_INFO + "] - " + Calendar.getInstance().getTime() + " - End of the game: you loose...");		
		
		for(int i=0; i<controls.length; i++)
		{
			// Check if there is a mine in that cell and show it.
			cell = (Cell)controls[i];
			
			if(cell.isButton())
			{
				cellInfo = (CellInfo)cell.getData();
				
				if(cellInfo.isMine() && !cellInfo.isFlag())
				{
					cell.setType(Constants.TYPE_LABEL);	
					cell.setImage(ResourceManager.getResourceManager().getImage(Constants.IMAGE_MINE));									
					cell.redraw();
				}
				else if(!cellInfo.isMine() && cellInfo.isFlag())
				{
					cell.setType(Constants.TYPE_LABEL);
					cell.setImage(ResourceManager.getResourceManager().getImage(Constants.IMAGE_WRONG));					
					cell.redraw();
				}
			}
			controls[i].removeMouseListener(this.cellListener);
		}
	}
	
	public Vector showAllCellsAround(CellInfo cellInfo)
	{
		// Variables.
		Vector list = new Vector();
		CellInfo aux = null;
		Control[] controls = minesField.getChildren();
		int columns = preferences.getColumns();
		
		// Algorithm
		if(cellInfo.isTopBorder())
		{
			if(cellInfo.isLeftBorder())
			{
				// The same line.
				aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
				list.addElement(aux);
					
				// The line below.
				aux = (CellInfo)controls[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
				list.addElement(aux);
					
				aux = (CellInfo)controls[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
				list.addElement(aux);
			}
			else if(cellInfo.isRightBorder())
			{
				// The same line.
				aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				list.addElement(aux);
				
				// The line below.
				aux = (CellInfo)controls[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y - 1].getData();
				list.addElement(aux);
					
				aux = (CellInfo)controls[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y].getData();
				list.addElement(aux);
			}
			else
			{
				// The same line.
				aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				list.addElement(aux);

				aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
				list.addElement(aux);
						
				// The line below.
				aux = (CellInfo)controls[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y - 1].getData();
				list.addElement(aux);

				aux = (CellInfo)controls[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
				list.addElement(aux);

				aux = (CellInfo)controls[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
				list.addElement(aux);
			}
		}
		else if(cellInfo.isDownBorder())
		{
			if(cellInfo.isLeftBorder())
			{
				// The same line.
				aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
				list.addElement(aux);
				
				// The line above.
				aux = (CellInfo)controls[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
				list.addElement(aux);

				aux = (CellInfo)controls[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
				list.addElement(aux);
			}
			else if(cellInfo.isRightBorder())
			{
				// The same line.
				aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				list.addElement(aux);
				
				// The line above.
				aux = (CellInfo)controls[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y - 1].getData();
				list.addElement(aux);

				aux = (CellInfo)controls[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y].getData();
				list.addElement(aux);
			}
			else
			{
				// The same line.
				aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				list.addElement(aux);

				aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
				list.addElement(aux);
					
				// The line above.
				aux = (CellInfo)controls[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y - 1].getData();
				list.addElement(aux);

				aux = (CellInfo)controls[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
				list.addElement(aux);

				aux = (CellInfo)controls[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
				list.addElement(aux);
			}
		}
		else if(cellInfo.isLeftBorder())
		{
			// The same line.
			aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
			list.addElement(aux);

			// The line above.
			aux = (CellInfo)controls[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
			list.addElement(aux);

			aux = (CellInfo)controls[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
			list.addElement(aux);

			// The line below.
			aux = (CellInfo)controls[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
			list.addElement(aux);

			aux = (CellInfo)controls[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
			list.addElement(aux);
		}
		else if(cellInfo.isRightBorder())
		{
			// The same line.
			aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
			list.addElement(aux);
						
			// The line above.
			aux = (CellInfo)controls[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
			list.addElement(aux);

			aux = (CellInfo)controls[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y - 1].getData();
			list.addElement(aux);

			// The line below.
			aux = (CellInfo)controls[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
			list.addElement(aux);

			aux = (CellInfo)controls[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y - 1].getData();
			list.addElement(aux);
		}
		else  // It doesn't touch any border.
		{
			// The same line.
			aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
			list.addElement(aux);

			aux = (CellInfo)controls[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
			list.addElement(aux);
						
			// The line above.
			aux = (CellInfo)controls[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
			list.addElement(aux);

			aux = (CellInfo)controls[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y - 1].getData();
			list.addElement(aux);

			aux = (CellInfo)controls[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
			list.addElement(aux);

			// The line below.
			aux = (CellInfo)controls[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
			list.addElement(aux);

			aux = (CellInfo)controls[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y - 1].getData();
			list.addElement(aux);

			aux = (CellInfo)controls[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
			list.addElement(aux);
		}

		return list;
	}
	
	public void showAllZeroWarningsAround(CellInfo cellInfo, Vector list)
	{
		// Check if that cell is already processed.
		if(list.contains(cellInfo))
			return;
					
		list.addElement(cellInfo);
		
		// Variables.
		CellInfo aux = null;
		Control[] c = minesField.getChildren();
		int columns = preferences.getColumns();
		
		// Algorithm
		if(cellInfo.isTopBorder())
		{
			if(cellInfo.isLeftBorder())
			{
				// The same line.
				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
				if(aux.getWarning() == 0 && !aux.isFlag())
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
					
				// The line below.
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
					
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
			}
			else if(cellInfo.isRightBorder())
			{
				// The same line.
				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
				
				// The line below.
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y - 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
					
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
			}
			else
			{
				// The same line.
				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);

				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
						
				// The line below.
				aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y - 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);

				aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);

				aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
			}
		}
		else if(cellInfo.isDownBorder())
		{
			if(cellInfo.isLeftBorder())
			{
				// The same line.
				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
				
				// The line above.
				aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);

				aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
			}
			else if(cellInfo.isRightBorder())
			{
				// The same line.
				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
				
				// The line above.
				aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y - 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);

				aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
			}
			else
			{
				// The same line.
				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);

				aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
					
				// The line above.
				aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y - 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);

				aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);

				aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
				if(aux.getWarning() == 0)
					this.showAllZeroWarningsAround(aux, list);
				else if(!list.contains(aux))
					list.addElement(aux);
			}
		}
		else if(cellInfo.isLeftBorder())
		{
			// The same line.
			aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			// The line above.
			aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			// The line below.
			aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);
		}
		else if(cellInfo.isRightBorder())
		{
			// The same line.
			aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);
						
			// The line above.
			aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y - 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			// The line below.
			aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y - 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);
		}
		else  // It doesn't touch any border.
		{
			// The same line.
			aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y - 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			aux = (CellInfo)c[cellInfo.getLocation().x * columns + cellInfo.getLocation().y + 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);
						
			// The line above.
			aux = (CellInfo)c[(cellInfo.getLocation().x - 1)* columns + cellInfo.getLocation().y].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y - 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			aux = (CellInfo)c[(cellInfo.getLocation().x - 1) * columns + cellInfo.getLocation().y + 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			// The line below.
			aux = (CellInfo)c[(cellInfo.getLocation().x + 1)* columns + cellInfo.getLocation().y].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y - 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);

			aux = (CellInfo)c[(cellInfo.getLocation().x + 1) * columns + cellInfo.getLocation().y + 1].getData();
			if(aux.getWarning() == 0)
				this.showAllZeroWarningsAround(aux, list);
			else if(!list.contains(aux))
				list.addElement(aux);
		}
	}

	public void finishGame()
	{
		boolean sw = true;
		CellInfo cellInfo = null;
		Control[] aux = minesField.getChildren();

		// Check if all mines has been selected.
		for(int i=0; i< aux.length; i++)
		{
			if(((Cell)aux[i]).isButton())
			{
				cellInfo = (CellInfo)aux[i].getData();
				if(cellInfo.isFlag())
				{
					if (!cellInfo.isMine())
						sw = false;
				}
				else 
					sw = false;
			}
		}		
		
		if(sw)
		{
			// Tracing.
			if(MinesweeperPlugin.getDefault().isDebugging())
				System.out.println("[Minesweeper - " + Constants.TRACE_INFO + "] - " + Calendar.getInstance().getTime() + " - End of the game: you win...");		

			view.getTimer().stop();
			view.getMainButton().setImage(ResourceManager.getResourceManager().getImage(Constants.IMAGE_HAPPY));

			for(int i=0; i< aux.length; i++)
				aux[i].removeMouseListener(this.cellListener);
			
			if(preferences.getMode().equals(Constants.MODE_EXPERT) && ((Integer)records.get(Constants.RECORD_TIME_MODE_EXPERT)).intValue() > view.getTimer().getTime())
				this.showNewRecordDialog();
			else if(preferences.getMode().equals(Constants.MODE_INTERMEDIATE) && ((Integer)records.get(Constants.RECORD_TIME_MODE_INTERMEDIATE)).intValue() > view.getTimer().getTime())
				this.showNewRecordDialog();
			else if(preferences.getMode().equals(Constants.MODE_BEGINNER) && ((Integer)records.get(Constants.RECORD_TIME_MODE_BEGINNER)).intValue() > view.getTimer().getTime())
				this.showNewRecordDialog();
		}
	}
	
	private void showNewRecordDialog()
	{
		// Show the new record dialog.
		NewRecordDialog dialog = new NewRecordDialog(view.getViewSite().getShell(), view);
		dialog.open();		
	}	
}